from __future__ import division, print_function
import csv
from cmath import exp, pi
import math
import numpy as np
import matplotlib.pyplot as plt
import scipy.linalg
np.set_printoptions(precision=4)
from control import *

import cmath
from scipy.optimize import leastsq, least_squares
from scipy.optimize import curve_fit



##ENTER YOUR FILE NAME HERE!  PLACE CSV FILE IN SAME DIRECTORY AS THIS PYTHON FILE
##CSV FILE MUST BE OF TYPE GENERATED FROM:
##Lab4_6302_1_sweeper code!
filename = "current.csv"


Tsamp = 1e-3

#set for noise purposes adjust up/down (most likely down) as needed
#Too low and there'll be nothing to fit though...
freq_limit = 5

#lists for original data storage:
frequency = []
mag1 = []
phase1 = []
mag2 = []
phase2 = []

#Extract CSV data initially:
with open(filename, 'rt') as csvfile:
    reader = csv.reader(csvfile)
    first = True
    for row in reader:
        if first:
            first = False
        else:
            if float(row[1])<freq_limit:
                frequency.append(float(row[1]))
                mag1.append(float(row[2]))
                phase1.append(float(row[3]))
                mag2.append(float(row[5]))
                phase2.append(float(row[6]))

#Fix phase discontinuities in overall phase
phase1 = [p-360 if p>20 else p for p in phase1]

#Group like frequency data points for averaging:
total_set = {}
for x in range(len(frequency)):
    total_set[frequency[x]] = total_set.get(frequency[x],[])
    total_set[frequency[x]].append([mag1[x],phase1[x],mag2[x],phase2[x]])

y_resp = []
u_resp =[]
mag1 = []
mag2=[]
phase1=[]
phase2=[]

#Average out frequency measurements:
#Create TF data of input-to-y and input-to-command/u
for x in sorted(list(total_set.keys())):
    temp = total_set[x]
    mag1avg = sum([x[0] for x in temp])/len(temp)
    phase1avg = sum([x[1] for x in temp])/len(temp)
    mag2avg = sum([x[2] for x in temp])/len(temp)
    phase2avg = sum([x[3] for x in temp])/len(temp)
    mag1.append(mag1avg)
    phase1.append(phase1avg)
    mag2.append(mag2avg)
    phase2.append(phase2avg)
    y_resp.append(mag1avg*exp(phase1avg*pi/180*1j))
    u_resp.append(mag2avg*exp(phase2avg*pi/180*1j))

#generate np array of frequency data
freq = np.array(sorted(list(total_set.keys())))
#convert to radial frequency:
omega = 2*pi*freq

#generate specific frequency data (mostly debug purposes)
mag = np.array(mag1)/np.array(mag2)
phase = np.array(phase1)-np.array(phase2)

#create data for fitting (CT transfer function of plant only)
final = np.array(y_resp)/np.array(u_resp)


####FITTING AREA:
##NOTE: Not a general transfer function fitter!
##Here we attempt to fit the frequency response data
##To a third-order system with three natural frequencies (poles)
##No feed-forward terms (zeros)
##Method is HIGHLY dependent on initial guess so will only generally work
##for our system as designed...if you make modifications a more advanced
##fitter will be needed.
##Fitting is done with scipy's leastsq (Levenberg-Marquardt algorithm)
##and due to nature of problem very prone to finding local mininima so
##adjustments from the default starting conditions should be done carefully

j = 1j

def residuals(p,y,x):
    n,d0,d1,d2=p
    s=j*x
    tf = n/(s**3+d2*(s**2)+d1*s+d0)
    diff=tf-y
    z1d = np.zeros(y.size*2, dtype = np.float64)
    z1d[0:z1d.size:2] = diff.real
    z1d[1:z1d.size:2] = diff.imag
    return z1d
  

###initial guess of C.T. natural frequencies @:
##-0.4
##-3.0 +/5j
p0=np.array([100,6,36,13])
res = leastsq(residuals,p0, args=(final,omega),maxfev=10000)
print("coefficients found:")
print(res)


####DISTRIBUTE FITTED PARAMETERS
##Data is used to construct a CT transfer function.
##That transfer function is then mapped to a CT state space model
##That CT state space model is then discretized using zoh approximation

n=res[0][0]
den = np.poly1d([1,res[0][3],res[0][2],res[0][1]])

#Build 
tf_sys = tf([n],den.c)
ss_sys = tf2ss(tf_sys)
ss_sys_d = matlab.c2d(ss_sys,Tsamp)

#State Space Model Generated:
print("State Space Model Generated:")
print(ss_sys_d)
A = ss_sys_d.A
B = ss_sys_d.B
C = ss_sys_d.C
D = ss_sys_d.D

print("Eigenvalues of State Space Model:")
print(ss_sys_d.pole())


#Pick Natural frequencies:
des_nat_freq = [0.99,0.99,0.99]
k_acker = acker(A,B,des_nat_freq)
print("K Gains for requested plant natural frequencies:")
print(k_acker)

#acker for observer picking:
des_obs_freq = [0.97,0.97,0.97]
l_acker = acker(A.T,C.T,des_obs_freq).T
print("L Gains for requested observer natural frequencies:")
print(l_acker)

k = k_acker #change to k_acker or manually set as needed
#k = k_lqr

#k = np.matrix([[3,1,20]])
#calculate K_u:
KuInv = C*np.linalg.inv(np.eye(3)-1*(A-B*k))*B
Ku = np.linalg.inv(KuInv)
print("Ku Value: %0.4f" %(Ku.item(0)))
Ac = A-B*k #equivalent to M in notes
Bc = B*Ku
Cc = C
Dc = D
#print(np.linalg.eig(Ac))

overall=ss(Ac,Bc,Cc,Dc,Tsamp)



maga, phasea,omegaa = bode(tf_sys, dB=True, deg=True, Hz=False)
mag_total =np.array(mag1)/np.array(mag2)
phase_total = np.array(phase1)-np.array(phase2)


t = np.arange(0,3,0.001)
sig = np.ones(np.size(t))
y,t1,x=lsim(overall,sig,t)

plot_in_radians = False
if plot_in_radians:
    scale = 1
else:
    scale = 180/3.14159
out = (C*x).A

####UNCOMMENT BELOW TO PLOT FREQUENCY DATA (MOSTLY DEBUG PURPOSES FOR US)
##command = np.array(Ku*sig[1] - k*x)[0]
##plt.subplot(5, 1, 1)
##x1, = plt.plot(x[0])
##plt.ylabel('x_1')
##plt.subplot(5, 1, 2)
##x2, = plt.plot(x[1])
##plt.ylabel('x_2')
##plt.subplot(5, 1, 3)
##x3, = plt.plot(x[2])
##plt.ylabel('x_3')
##plt.subplot(5, 1, 4)
##o, = plt.plot(out[0])
##plt.ylabel('output')
##plt.subplot(5, 1, 5)
##com, = plt.plot(command)
##plt.ylabel('command')
##plt.xlabel('time steps')
##plt.show()


####UNCOMMENT BELOW IF YOU'D LIKE TO PLOT SS RESPONSE 
##plt.subplot(4, 1, 1)
###plt.semilogx(omega, [1.0/x for x in mag])
##plt.semilogx(omegaa, maga,omega, [20*math.log(x,10) for x in mag_total],'r')
##plt.ylabel('Mag1')
##plt.subplot(4, 1,2)
###plt.semilogx(omega, [1.0/fit_itf(x) for x in omega])
##plt.semilogx(omegaa, phasea, omega, phase_total, 'r')
##plt.ylabel('Phase1')
##plt.subplot(4, 1, 3)
###plt.semilogx(omegaa, maga)
##plt.semilogx(omega, [20*math.log(x,10) for x in abs(final)])
##plt.ylabel('Mag2')
##plt.subplot(4, 1, 4)
##plt.semilogx(omega, 180/3.14159*np.angle(final))
###plt.semilogx(omegaa, phasea)
##plt.ylabel('Phase2')
##plt.xlabel('frequency (Hz)')
##plt.show()
